#
# Include this module to enable the NamedParameters#named_parameters macros
#
module NamedParameters

  # Works like NamedParameters#named_parameters, but instead of using it right before a method definition,
  # you use it afterwards. You can also add the strict and super mode. Replace the +my_method+ with the name of the
  # method that you want to have named parameters
  # @see NamedParameters#named_parameters
  # @see NamedParameters#named_parameters_strict
  # @see NamedParameters#named_parameters_super
  # @see NamedParameters#named_parameters_strict_super
  def my_method_has_named_parameters(optionals = { })
    # Empty method just for the documentation. Those methods are realized in NamedParameters#method_missing
  end

  # Makes the next method callable via named parameters (options hash)
  # The caller hash may contain keys that are not contained in the parameters of the next method
  # @param [Hash[Symbol,Object]] optionals Optional default values for the parameters
  def named_parameters(optionals = { })
    _named_parameters optionals, false, false
  end

  # Makes the next method callable via named parameters (options hash)
  # The caller hash may only contain keys that are in the parameters of the next method
  # @param [Hash[Symbol,Object]] optionals Optional default values for the parameters
  def named_parameters_strict(optionals = { })
    _named_parameters optionals, true, false
  end

  # Makes the next method callable via named parameters (options hash) and calls super with this options hash as the
  # only parameter
  # The caller hash may contain keys that are not contained in the parameters of the next method
  # @param [Hash[Symbol,Object]] optionals Optional default values for the parameters
  def named_parameters_super(optionals = { })
    _named_parameters optionals, false, true
  end

  # Makes the next method callable via named parameters (options hash) and calls super with this options hash as the
  # only parameter
  # The caller hash may only contain keys that are in the parameters of the next method
  # @param [Hash[Symbol,Object]] optionals Optional default values for the parameters
  def named_parameters_strict_super(optionals = { })
    _named_parameters optionals, true, true
  end

  alias_method :named_parameters_super_strict, :named_parameters_strict_super

  private

  # Makes all methods of this module to class or module methods of include
  def self.included base #:nodoc:
    base.extend(self)
  end

  #noinspection RubyArgCount
  self.class.class_eval do
    alias :method_added_before_named_parameters :method_added
    private :method_added_before_named_parameters
  end

  # Enables the <method>_has_named_parameter ghost methods
  def method_missing method, *args, &block
    # catch only methods like <method_name> + _has_named_parameters + <mode definition>
    if /\A(?<method_name>.+)_has_named_parameters(?<mode>_super_strict|_strict_super|_super|_strict)?\z/ =~ method.to_s
      # find out mode
      mode = "" if mode.nil?
      call_super = mode.include? 'super'
      # find out strict
      strict     = mode.include? 'strict'
      # redefine
      _redefine_with_named_parameters(method_name,
                                      args[0] || { },
                                      strict,
                                      call_super)
    else
      # if method is not like above pattern, do the original stuff
      super method, *args, &block
    end
  end

  # Makes sure a method is redefined as soon as it is defined. Only when named_parameters was called right before
  def method_added(method_name)
    # redefine the method
    if @redefine_next
      @redefine_next = false
      _redefine_with_named_parameters(method_name,
                                      @named_parameters_optionals,
                                      @named_parameters_strict,
                                      @named_parameters_super)
    end
    # call old method_added
    method_added_before_named_parameters method_name
  end

  # @param [Hash[Symbol,Object]] optionals Optional default values for the parameters
  # @param [Boolean] strict Shall keys in the caller options hash, that are not in the parameter list, throw an error?
  # @param [Boolean] call_super Shall super be called with the caller option hash as the only parameter?
  def _named_parameters(optionals, strict, call_super)
    # remember the preferences for the next method, that is added
    @named_parameters_optionals = optionals
    @named_parameters_strict    = strict
    @named_parameters_super     = call_super
    # make sure only one method is redefined after the named_parameters macro was called
    @redefine_next              = true
  end

  # Redefines a method so the caller can use named parameters instead of ordinary parameters
  # @param [Symbol] method_name
  # @param [Hash[Symbol,Object]] optionals Optional default values for the parameters
  # @param [Boolean] strict Shall keys in the caller options hash, that are not in the parameter list, throw an error?
  # @param [Boolean] call_super Shall super be called with the caller option hash as the only parameter?
  def _redefine_with_named_parameters(method_name, optionals, strict, call_super)
    # the method as defined in the class/module
    original_method              ||= instance_method method_name
    # names of the parameters as defined in the class/module
    parameter_names              ||= original_method.parameters.collect { |_, b| b }
    new_name_for_original_method ||= (method_name.to_s + '_before_named_parameters').to_sym

    # "safe" original method
    alias_method new_name_for_original_method, method_name

    # redefine old method with new logic redirecting to original method
    define_method method_name do |options = { }, &block|
      # check parameters if strict
      _check_for_not_needed_options options, parameter_names, method_name if strict
      # call super if so wanted
      super(options) if call_super

      # build parameter array from options and optionals
      parameters ||= _parameters_from_options options, optionals, parameter_names, method_name

      # redirect to original method
      send new_name_for_original_method, *parameters, &block
    end
  end

  # Builds the actual parameter array for the redirected call to the original method
  def _parameters_from_options(options, optionals, parameter_names, method_name)
    parameters ||= []
    parameter_names.each do |name|
      parameters << options.fetch(name) { optionals.fetch(name) { raise "Mandatory parameter #{name} is not passed to
method #{method_name}" } }
    end
    parameters
  end

  # Checks whether there are keys in the options, that are not in the parameter_names. Throws a user friendly
  # RuntimeError in that case
  def _check_for_not_needed_options(options, parameter_names, method_name)
    options.each do |name, _|
      raise ("Parameter '#{name}' is not accepted by '#{method_name}'") unless parameter_names.member? name
    end
  end

end
